@@ -118,7 +118,7 @@ int pv_main_loop(pvstate_t state)
 	 * don't need to.
 	 */
 
-	pv_crs_init(state);
+	pv_crs_init(&(state->cursor), &(state->control), &(state->flags));
 
 	eof_in = false;
 	eof_out = false;
@@ -166,7 +166,7 @@ int pv_main_loop(pvstate_t state)
 	 */
 	if (input_fd < 0) {
 		if (state->control.cursor)
-			pv_crs_fini(state);
+			pv_crs_fini(&(state->cursor), &(state->control), &(state->flags));
 		return state->status.exit_status;
 	}
 #if HAVE_POSIX_FADVISE
@@ -219,11 +219,11 @@ int pv_main_loop(pvstate_t state)
 		 * Check for remote messages from -R every short while.
 		 */
 		if (pv_elapsedtime_compare(&cur_time, &next_remotecheck) > 0) {
-			pv_remote_check(state);
+			(void) pv_remote_check(state);
 			pv_elapsedtime_add_nsec(&next_remotecheck, REMOTE_INTERVAL);
 		}
 
-		if (1 == state->flag.trigger_exit)
+		if (1 == state->flags.trigger_exit)
 			break;
 
 		if (state->control.rate_limit > 0) {
@@ -270,7 +270,7 @@ int pv_main_loop(pvstate_t state)
 		if (written < 0) {
 			debug("%s: %s", "write error from pv_transfer", strerror(errno));
 			if (state->control.cursor)
-				pv_crs_fini(state);
+				pv_crs_fini(&(state->cursor), &(state->control), &(state->flags));
 			return state->status.exit_status;
 		}
 
@@ -292,7 +292,7 @@ int pv_main_loop(pvstate_t state)
 		if (output_is_pipe) {
 			int nbytes;
 			nbytes = 0;
-			if (0 != state->flag.pipe_closed) {
+			if (0 != state->flags.pipe_closed) {
 				if (0 != state->transfer.written_but_not_consumed)
 					debug("%s",
 					      "clearing written_but_not_consumed because the output pipe was closed");
@@ -394,7 +394,7 @@ int pv_main_loop(pvstate_t state)
 		 */
 		if (eof_in && eof_out && 0 == state->transfer.written_but_not_consumed) {
 			final_update = true;
-			if ((state->display.display_visible)
+			if ((state->display.output_produced)
 			    || (state->control.delay_start < 0.001)) {
 				pv_elapsedtime_copy(&next_update, &cur_time);
 			}
@@ -487,10 +487,10 @@ int pv_main_loop(pvstate_t state)
 		state->transfer.elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
 
 		/* Resize the display, if a resize signal was received. */
-		if (1 == state->flag.terminal_resized) {
+		if (1 == state->flags.terminal_resized) {
 			unsigned int new_width, new_height;
 
-			state->flag.terminal_resized = 0;
+			state->flags.terminal_resized = 0;
 
 			new_width = (unsigned int) (state->control.width);
 			new_height = state->control.height;
@@ -506,10 +506,13 @@ int pv_main_loop(pvstate_t state)
 
 		if (state->control.no_display) {
 			/* If there's no display, calculate rate for the statistics. */
-			pv_calculate_transfer_rate(state, final_update);
+			pv_calculate_transfer_rate(&(state->calc), &(state->transfer), &(state->control),
+						   &(state->display), final_update);
 		} else {
 			/* Produce the display. */
-			pv_display(state, final_update);
+			pv_display(&(state->status), &(state->control), &(state->flags), &(state->transfer),
+				   &(state->calc), &(state->cursor), &(state->display), &(state->extra_display),
+				   final_update);
 		}
 	}
 
@@ -517,14 +520,14 @@ int pv_main_loop(pvstate_t state)
 	      eof_out ? "true" : "false");
 
 	if (state->control.cursor) {
-		pv_crs_fini(state);
+		pv_crs_fini(&(state->cursor), &(state->control), &(state->flags));
 	} else {
 		if ((!state->control.numeric) && (!state->control.no_display)
-		    && (state->display.display_visible))
-			pv_tty_write(state, "\n", 1);
+		    && (state->display.output_produced))
+			pv_tty_write(&(state->flags), "\n", 1);
 	}
 
-	if (1 == state->flag.trigger_exit)
+	if (1 == state->flags.trigger_exit)
 		state->status.exit_status |= PV_ERROREXIT_SIGNAL;
 
 	if (input_fd >= 0)
@@ -562,7 +565,7 @@ int pv_main_loop(pvstate_t state)
 				rate_deviation, state->control.bits ? _("b/s") : _("B/s"));
 
 		if (stats_size > 0 && stats_size < (int) (sizeof(stats_buf)))
-			pv_tty_write(state, stats_buf, (size_t) stats_size);
+			pv_tty_write(&(state->flags), stats_buf, (size_t) stats_size);
 	} else if (state->control.show_stats && state->calc.measurements_taken < 1) {
 		char msg_buf[256];	 /* flawfinder: ignore */
 		int msg_size;
@@ -573,7 +576,7 @@ int pv_main_loop(pvstate_t state)
 		msg_size = pv_snprintf(msg_buf, sizeof(msg_buf), "%s\n", _("rate not measured"));
 
 		if (msg_size > 0 && msg_size < (int) (sizeof(msg_buf)))
-			pv_tty_write(state, msg_buf, (size_t) msg_size);
+			pv_tty_write(&(state->flags), msg_buf, (size_t) msg_size);
 	}
 
 	return state->status.exit_status;
@@ -581,193 +584,60 @@ int pv_main_loop(pvstate_t state)
 
 
 /*
- * Watch the progress of file descriptor state->control.watch_fd in process
- * state->control.watch_pid and show details about the transfer on standard error
- * according to the given options.
- *
- * Returns nonzero on error.
+ * Return true if the given string contains a "%N" or "%{name}" format
+ * sequence.
  */
-int pv_watchfd_loop(pvstate_t state)
+static bool format_contains_name(const char *format_string)
 {
-	struct pvwatchfd_s info;
-	off_t position_now;
-	struct timespec next_update, cur_time;
-	struct timespec init_time, next_remotecheck, transfer_elapsed;
-	bool ended, first_check;
-	int rc;
-
-	memset(&info, 0, sizeof(info));
-	info.watch_pid = state->control.watch_pid;
-	info.watch_fd = state->control.watch_fd;
-	rc = pv_watchfd_info(state, &info, false);
-	if (0 != rc) {
-		state->status.exit_status |= PV_ERROREXIT_ACCESS;
-		/*@-compdestroy@ */
-		return state->status.exit_status;
-		/*@+compdestroy@ */
-		/* splint: no leak of info.state as it is unused so far. */
-	}
-
-	/*
-	 * Use a size if one was passed, otherwise use the total size
-	 * calculated.
-	 */
-	if (0 >= state->control.size)
-		state->control.size = info.size;
-
-	if (state->control.size < 1) {
-		char *fmt;
-		while (NULL != (fmt = strstr(state->control.default_format, "%e"))) {
-			debug("%s", "zero size - removing ETA");
-			/* strlen-1 here to include trailing \0 */
-			memmove(fmt, fmt + 2, strlen(fmt) - 1);	/* flawfinder: ignore */
-			/* flawfinder: default_format is always \0 terminated */
-			state->flag.reparse_display = 1;
-		}
-	}
-
-	memset(&cur_time, 0, sizeof(cur_time));
-	memset(&next_remotecheck, 0, sizeof(next_remotecheck));
-	memset(&next_update, 0, sizeof(next_update));
-	memset(&init_time, 0, sizeof(init_time));
-	memset(&transfer_elapsed, 0, sizeof(transfer_elapsed));
-
-	pv_elapsedtime_read(&cur_time);
-	pv_elapsedtime_copy(&(info.start_time), &cur_time);
-	pv_elapsedtime_copy(&next_remotecheck, &cur_time);
-	pv_elapsedtime_copy(&next_update, &cur_time);
-	pv_elapsedtime_add_nsec(&next_update, (long long) (1000000000.0 * state->control.interval));
-
-	ended = false;
-	first_check = true;
-
-	while (!ended) {
-		/*
-		 * Check for remote messages from -R every short while.
-		 */
-		if (pv_elapsedtime_compare(&cur_time, &next_remotecheck) > 0) {
-			pv_remote_check(state);
-			pv_elapsedtime_add_nsec(&next_remotecheck, REMOTE_INTERVAL);
-		}
-
-		if (1 == state->flag.trigger_exit)
-			break;
-
-		position_now = pv_watchfd_position(&info);
-
-		if (position_now < 0) {
-			ended = true;
-		} else {
-			if (first_check) {
-				state->display.initial_offset = position_now;
-				first_check = false;
-			}
-			state->transfer.transferred = position_now;
-			state->transfer.total_written = position_now;
-		}
-
-		pv_elapsedtime_read(&cur_time);
-
-		/* Ended - force a display update. */
-		if (ended) {
-			pv_elapsedtime_copy(&next_update, &cur_time);
-		}
-
-		/*
-		 * Restart the loop after a brief delay, if it's not time to
-		 * update the display.
-		 */
-		if (pv_elapsedtime_compare(&cur_time, &next_update) < 0) {
-			pv_nanosleep(50000000);
+	while ('\0' != format_string[0]) {
+		/* If we're not on a '%' character, move on. */
+		if ('%' != format_string[0]) {
+			format_string++;
 			continue;
 		}
-
-		pv_elapsedtime_add_nsec(&next_update, (long long) (1000000000.0 * state->control.interval));
-
-		/* Set the "next update" time to now, if it's in the past. */
-		if (pv_elapsedtime_compare(&next_update, &cur_time) < 0)
-			pv_elapsedtime_copy(&next_update, &cur_time);
-
-		/*
-		 * Calculate the effective start time: the time we actually
-		 * started, plus the total time we spent stopped.
-		 */
-		pv_elapsedtime_add(&init_time, &(info.start_time), &(state->signal.toffset));
-
-		/*
-		 * Now get the effective elapsed transfer time - current
-		 * time minus effective start time.
-		 */
-		pv_elapsedtime_subtract(&transfer_elapsed, &cur_time, &init_time);
-
-		state->transfer.elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
-
-		/* Resize the display, if a resize signal was received. */
-		if (1 == state->flag.terminal_resized) {
-			unsigned int new_width, new_height;
-
-			state->flag.terminal_resized = 0;
-
-			new_width = (unsigned int) (state->control.width);
-			new_height = state->control.height;
-			pv_screensize(&new_width, &new_height);
-
-			if (new_width > PVDISPLAY_WIDTH_MAX)
-				new_width = PVDISPLAY_WIDTH_MAX;
-
-			if (!state->control.width_set_manually)
-				state->control.width = (pvdisplay_width_t) new_width;
-			if (!state->control.height_set_manually)
-				state->control.height = new_height;
+		/* Move past the '%'. */
+		format_string++;
+		/* Early return if we've run out of string. */
+		if ('\0' == format_string[0])
+			return false;
+		/* If we've found "%N", return true. */
+		if ('N' == format_string[0])
+			return true;
+		/* If we've found "%%", move on and go back round the loop. */
+		if ('%' == format_string[0]) {
+			format_string++;
+			continue;
 		}
-
-		pv_display(state, ended);
+		/* Move past any digits, for the "%123{name}" syntax. */
+		while ('\0' != format_string[0] && (format_string[0] >= '0' && format_string[0] <= '9'))
+			format_string++;
+		/*
+		 * If the next character isn't '{', this isn't a valid
+		 * "%{xyz}" sequence, so go back to the top of the loop to
+		 * skip to the next '%'.
+		 */
+		if ('{' != format_string[0])
+			continue;
+		/*
+		 * If the next part is "{name}", forming "%{name}", return
+		 * true.
+		 */
+		if (0 == strncmp(format_string, "{name}", 6))
+			return true;
+		/* Loop round again to keep looking. */
 	}
-
-	if (!state->control.numeric)
-		pv_tty_write(state, "\n", 1);
-
-	if (1 == state->flag.trigger_exit)
-		state->status.exit_status |= PV_ERROREXIT_SIGNAL;
-
-	/*
-	 * Free the state structure specific to this file descriptor. 
-	 * Uunused so far - this is so that in future we could watch
-	 * multiple file descriptors similar to pv_watchpid_loop().
-	 */
-	if (NULL != info.state) {
-		pv_state_free(info.state);
-		info.state = NULL;
-	}
-
-	/*@-compdestroy@ */
-	return state->status.exit_status;
-	/*@+compdestroy@ */
-	/* splint: no leak of info.state as we just freed it above. */
+	return false;
 }
 
 
 /*
- * Watch the progress of all file descriptors in process state->control.watch_pid
- * and show details about the transfers on standard error according to the
- * given options.
- *
- * Replaces format_string in "state" so that starts with "%N " if it doesn't
- * already do so.
- *
- * Returns nonzero on error.
+ * Update the main format string for use with --watchfd, such that if
+ * necessary, it starts with "%N " if it doesn't already.
  */
-int pv_watchpid_loop(pvstate_t state)
+static void pv_watchfd_update_format_string(pvstate_t state)
 {
 	const char *original_format_string;
 	char new_format_string[512];	 /* flawfinder: ignore */
-	struct pvwatchfd_s *info_array = NULL;
-	int array_length = 0;
-	int fd_to_idx[FD_SETSIZE];
-	struct timespec next_update, cur_time;
-	int idx;
-	int prev_displayed_lines, blank_lines;
-	bool first_pass = true;
 
 	/*
 	 * flawfinder rationale (new_format_string): zeroed with memset(),
@@ -775,71 +645,236 @@ int pv_watchpid_loop(pvstate_t state)
 	 * explicitly terminated with \0.
 	 */
 
-	/*
-	 * Make sure the process exists first, so we can give an error if
-	 * it's not there at the start.
-	 */
-	if (kill(state->control.watch_pid, 0) != 0) {
-		pv_error(state, "%s %u: %s", _("pid"), state->control.watch_pid, strerror(errno));
-		state->status.exit_status |= PV_ERROREXIT_ACCESS;
-		return PV_ERROREXIT_ACCESS;
-	}
+	/* If there's nothing to watch, do nothing. */
+	if (state->watchfd.count < 1)
+		return;
+	if (NULL == state->watchfd.pid)
+		return;
+	if (NULL == state->watchfd.fd)
+		return;
 
 	/*
 	 * Make sure there's a format string, and then insert %N into it if
-	 * it's not present.
+	 * it's not present AND we're either watching more than one item, or
+	 * the item we're watching is a whole PID, not a single FD.
 	 */
-	original_format_string =
-	    NULL != state->control.format_string ? state->control.format_string : state->control.default_format;
+	original_format_string = state->control.format_string;
+	if (NULL == original_format_string)
+		original_format_string = state->control.default_format;
+	if (NULL == original_format_string)
+		return;
+
 	memset(new_format_string, 0, sizeof(new_format_string));
-	if (NULL == original_format_string) {
-		(void) pv_snprintf(new_format_string, sizeof(new_format_string), "%%N");
-	} else if (NULL == strstr(original_format_string, "%N")) {
-		(void) pv_snprintf(new_format_string, sizeof(new_format_string), "%%N %s", original_format_string);
+
+	if (state->watchfd.count > 1 || -1 == state->watchfd.fd[0]) {
+		/* Watching more than one single FD; need %N. */
+		if (!format_contains_name(original_format_string)) {
+			(void) pv_snprintf(new_format_string, sizeof(new_format_string), "%%N %s",
+					   original_format_string);
+		} else {
+			/* If the format string is set, nothing to do. */
+			if (NULL != state->control.format_string)
+				return;
+			/* Prepare a copy of the default format. */
+			(void) pv_snprintf(new_format_string, sizeof(new_format_string), "%s", original_format_string);
+		}
 	} else {
+		/* Watching only one FD; don't prepend %N. */
+		/* If the format string is set, nothing to do. */
+		if (NULL != state->control.format_string)
+			return;
+		/* Prepare a copy of the default format. */
 		(void) pv_snprintf(new_format_string, sizeof(new_format_string), "%s", original_format_string);
 	}
+
 	new_format_string[sizeof(new_format_string) - 1] = '\0';
+
 	if (NULL != state->control.format_string)
 		free(state->control.format_string);
 	state->control.format_string = pv_strdup(new_format_string);
+}
+
+
+/*
+ * Watch the progress of the PID:FD pairs, or of all FDs under PIDs, as
+ * specified in state->watchfd, showing details on standard error according
+ * to the given options.
+ *
+ * Replaces format_string in "state" so that starts with "%N " if it doesn't
+ * already do so.
+ *
+ * Returns nonzero on error.
+ */
+int pv_watchfd_loop(pvstate_t state)
+{
+	struct {			 /* details for each PID:FD specified by operator */
+		pid_t pid;		 /* watched PID */
+		int fd;			 /* watched fd, or -1 for all */
+		pvwatchfd_t info_array;	 /* watch information for each fd */
+		int array_length;	 /* length of watch info array */
+		bool finished;		 /* "PID:FD": fd closed; or PID gone */
+	} *watching;
+	unsigned int watch_idx;
+	bool all_watching_finished;
+	struct timespec next_update, next_remotecheck, cur_time;
+	int prev_displayed_lines, blank_lines;
 
 	/*
-	 * Get things ready for the main loop.
+	 * In each watching[].info_array[] fd info item, once its "watch_fd"
+	 * is closed, its "closed" flag is set to true and its "end_time" to
+	 * now.  It isn't marked "unused" until its "end_time" is long
+	 * enough ago, so the information is held on-screen for a short
+	 * while after the fd closes or the PID exits (#81).
+	 */
+
+	/* If there's nothing to watch, do nothing at all. */
+	if (state->watchfd.count < 1)
+		return 0;
+	if (NULL == state->watchfd.pid)
+		return PV_ERROREXIT_MEMORY;
+	if (NULL == state->watchfd.fd)
+		return PV_ERROREXIT_MEMORY;
+
+	/*
+	 * Make sure there's no name set.
+	 */
+	if (NULL != state->control.name) {
+		free(state->control.name);
+		state->control.name = NULL;
+	}
+
+	/* Adjust the format string so PID:FD or FD prefixes are shown. */
+	pv_watchfd_update_format_string(state);
+
+	/*
+	 * Allocate the watching array.
+	 */
+	watching = malloc(state->watchfd.count * sizeof(*watching));
+	if (NULL == watching) {
+		pv_error("%s: %s", _("buffer allocation failed"), strerror(errno));
+		return PV_ERROREXIT_MEMORY;
+	}
+	memset(watching, 0, state->watchfd.count * sizeof(*watching));
+
+	/*
+	 * Populate the watching array.  In the process of doing so, raise
+	 * errors if any of the initially specified PIDs or PID:FD pairs
+	 * don't exist or aren't readable.
+	 */
+	for (watch_idx = 0; watch_idx < state->watchfd.count; watch_idx++) {
+		int rc;
+
+		watching[watch_idx].pid = state->watchfd.pid[watch_idx];
+		watching[watch_idx].fd = state->watchfd.fd[watch_idx];
+		watching[watch_idx].info_array = NULL;
+		watching[watch_idx].array_length = 0;
+		watching[watch_idx].finished = false;
+
+		if (kill(watching[watch_idx].pid, 0) != 0) {
+			/* Inaccessible PID - error, mark as finished. */
+			pv_error("%s %u: %s", _("pid"), watching[watch_idx].pid, strerror(errno));
+			state->status.exit_status |= PV_ERROREXIT_ACCESS;
+			watching[watch_idx].finished = true;
+			continue;
+		}
+
+		if (-1 == watching[watch_idx].fd) {
+			/*
+			 * If this is a whole-PID watch, do the initial FD
+			 * scan to list all the FDs under that PID.
+			 */
+			rc = pv_watchpid_scanfds(state, watching[watch_idx].pid, -1,
+						 &(watching[watch_idx].array_length),
+						 &(watching[watch_idx].info_array));
+			if (rc != 0) {
+				/* Scan failed - error, mark as finished. */
+				pv_error("%s %u: %s", _("pid"), watching[watch_idx].pid, strerror(errno));
+				state->status.exit_status |= PV_ERROREXIT_ACCESS;
+				watching[watch_idx].finished = true;
+			}
+		} else {
+			/*
+			 * Scan the PID for the one specific FD given.
+			 */
+			rc = pv_watchpid_scanfds(state, watching[watch_idx].pid, watching[watch_idx].fd,
+						 &(watching[watch_idx].array_length),
+						 &(watching[watch_idx].info_array));
+			if (rc != 0) {
+				/* Scan failed - mark as finished. */
+				state->status.exit_status |= PV_ERROREXIT_ACCESS;
+				watching[watch_idx].finished = true;
+			} else if (0 == watching[watch_idx].array_length) {
+				/* FD not found - error, mark as finished. */
+				pv_error("%s %u: %s %d: %s",
+					 _("pid"), watching[watch_idx].pid, _("fd"), watching[watch_idx].fd,
+					 strerror(ENOENT));
+				state->status.exit_status |= PV_ERROREXIT_ACCESS;
+				watching[watch_idx].finished = true;
+			} else if (!watching[watch_idx].info_array[0].displayable) {
+				/* FD not displayable - mark as finished. */
+				state->status.exit_status |= PV_ERROREXIT_ACCESS;
+				watching[watch_idx].finished = true;
+			}
+		}
+	}
+
+	/*
+	 * Skip to the end if none of the items to watch can be watched.
+	 */
+	all_watching_finished = true;
+	for (watch_idx = 0; watch_idx < state->watchfd.count; watch_idx++) {
+		if (watching[watch_idx].finished)
+			continue;
+		all_watching_finished = false;
+		break;
+	}
+	if (all_watching_finished)
+		goto end_pv_watchfd_loop;
+
+	/*
+	 * Prepare timing structures for the main loop.
 	 */
 
 	memset(&cur_time, 0, sizeof(cur_time));
+	memset(&next_remotecheck, 0, sizeof(next_remotecheck));
 	memset(&next_update, 0, sizeof(next_update));
 
 	pv_elapsedtime_read(&cur_time);
+	pv_elapsedtime_copy(&next_remotecheck, &cur_time);
 	pv_elapsedtime_copy(&next_update, &cur_time);
 	pv_elapsedtime_add_nsec(&next_update, (long long) (1000000000.0 * state->control.interval));
 
-	for (idx = 0; idx < FD_SETSIZE; idx++) {
-		fd_to_idx[idx] = -1;
-	}
+	/*
+	 * Main loop - continually check each watching[] item and display
+	 * progress until all watched items have finished or an exit signal
+	 * is received.
+	 */
 
 	prev_displayed_lines = 0;
+	all_watching_finished = false;
 
-	while (true) {
-		int rc, fd, displayed_lines;
+	while (!all_watching_finished) {
+		int displayed_lines;
+		bool terminal_resized;
 
-		if (1 == state->flag.trigger_exit)
-			break;
-
-		pv_elapsedtime_read(&cur_time);
-
-		if (kill(state->control.watch_pid, 0) != 0) {
-			if (first_pass) {
-				pv_error(state, "%s %u: %s", _("pid"), state->control.watch_pid, strerror(errno));
-				state->status.exit_status |= PV_ERROREXIT_ACCESS;
-				if (NULL != info_array)
-					free(info_array);
-				return PV_ERROREXIT_ACCESS;
+		/* Check for remote messages from -R every short while. */
+		if (pv_elapsedtime_compare(&cur_time, &next_remotecheck) > 0) {
+			if (pv_remote_check(state)) {
+				/* Message received - ensure format has %N. */
+				pv_watchfd_update_format_string(state);
+				/* Fake a resize, to force a reparse. */
+				state->flags.terminal_resized = 1;
 			}
-			break;
+			pv_elapsedtime_add_nsec(&next_remotecheck, REMOTE_INTERVAL);
 		}
 
+		/* End the loop if a signal handler has set the exit flag. */
+		if (1 == state->flags.trigger_exit)
+			break;
+
+		/* Get the current time. */
+		pv_elapsedtime_read(&cur_time);
+
 		/*
 		 * Restart the loop after a brief delay, if it's not time to
 		 * update the display.
@@ -849,17 +884,25 @@ int pv_watchpid_loop(pvstate_t state)
 			continue;
 		}
 
+		/* Set the time of the next display update. */
 		pv_elapsedtime_add_nsec(&next_update, (long long) (1000000000.0 * state->control.interval));
 
 		/* Set the "next update" time to now, if it's in the past. */
 		if (pv_elapsedtime_compare(&next_update, &cur_time) < 0)
 			pv_elapsedtime_copy(&next_update, &cur_time);
 
-		/* Resize the display, if a resize signal was received. */
-		if (1 == state->flag.terminal_resized) {
+		/*
+		 * Resize the display, if a resize signal was received.
+		 *
+		 * We also set a local flag to tell the display loop below
+		 * to trigger a name reset and display reparse for every FD
+		 * output line, to inherit the size change.
+		 */
+		terminal_resized = false;
+		if (1 == state->flags.terminal_resized) {
 			unsigned int new_width, new_height;
 
-			state->flag.terminal_resized = 0;
+			state->flags.terminal_resized = 0;
 
 			new_width = (unsigned int) (state->control.width);
 			new_height = state->control.height;
@@ -871,124 +914,208 @@ int pv_watchpid_loop(pvstate_t state)
 			state->control.width = (pvdisplay_width_t) new_width;
 			state->control.height = new_height;
 
-			for (idx = 0; NULL != info_array && idx < array_length; idx++) {
-				if (NULL == info_array[idx].state)
-					continue;
-				info_array[idx].state->control.width = state->control.width;
-				info_array[idx].state->control.height = state->control.height;
-				pv_watchpid_setname(state, &(info_array[idx]));
-				info_array[idx].state->flag.reparse_display = 1;
-			}
+			terminal_resized = true;
 		}
 
-		rc = pv_watchpid_scanfds(state, state->control.watch_pid, &array_length, &info_array, fd_to_idx);
-		if (rc != 0) {
-			if (first_pass) {
-				pv_error(state, "%s %u: %s", _("pid"), state->control.watch_pid, strerror(errno));
-				state->status.exit_status |= PV_ERROREXIT_ACCESS;
-				if (NULL != info_array)
-					free(info_array);
-				return PV_ERROREXIT_ACCESS;
-			}
-			break;
-		}
-
-		first_pass = false;
+		/*
+		 * Run through each watched item.
+		 */
 		displayed_lines = 0;
+		for (watch_idx = 0; NULL != watching && watch_idx < state->watchfd.count; watch_idx++) {
+			int info_idx;
 
-		for (fd = 0; fd < FD_SETSIZE && NULL != info_array; fd++) {
-			off_t position_now;
-			struct timespec init_time, transfer_elapsed;
-
-			if (displayed_lines >= (int) (state->control.height))
-				break;
-
-			idx = fd_to_idx[fd];
-
-			if (idx < 0)
+			/* Skip watched items that have finished. */
+			if (watching[watch_idx].finished)
 				continue;
 
-			if (info_array[idx].watch_fd < 0) {
+			if (-1 == watching[watch_idx].fd) {
+				int rc;
 				/*
-				 * Non-displayable fd - just remove if
-				 * changed
+				 * If this watched item is a whole PID,
+				 * rescan that PID's FDs.
 				 */
-				if (pv_watchfd_changed(&(info_array[idx]))) {
-					fd_to_idx[fd] = -1;
-					info_array[idx].watch_pid = 0;
-					if (NULL != info_array[idx].state)
-						pv_state_free(info_array[idx].state);
-					/*@-mustfreeonly@ *//* not a leak - we've just free()d it. */
-					info_array[idx].state = NULL;
-					/*@+mustfreeonly@ */
-					debug("%s %d: %s", "fd", fd, "removing");
+				rc = pv_watchpid_scanfds(state, watching[watch_idx].pid, -1,
+							 &(watching[watch_idx].array_length),
+							 &(watching[watch_idx].info_array));
+				if (rc != 0) {
+					/*
+					 * PID now inaccessible - mark it as
+					 * finished.
+					 */
+					watching[watch_idx].finished = true;
 				}
-				continue;
-			}
-
-			if (NULL == info_array[idx].state) {
-				debug("%s %d: %s", "fd", fd, "null state - skipping");
-				continue;
-			}
-
-			/*
-			 * Displayable fd - display, or remove if changed
-			 */
-
-			position_now = pv_watchfd_position(&(info_array[idx]));
-
-			if (position_now < 0) {
-				fd_to_idx[fd] = -1;
-				info_array[idx].watch_pid = 0;
-				if (NULL != info_array[idx].state)
-					pv_state_free(info_array[idx].state);
-				/*@-mustfreeonly@ *//* not a leak - we've just free()d it. */
-				info_array[idx].state = NULL;
-				/*@+mustfreeonly@ */
-				debug("%s %d: %s", "fd", fd, "removing");
-				continue;
-			}
-
-			info_array[idx].position = position_now;
-
-			memset(&init_time, 0, sizeof(init_time));
-			memset(&transfer_elapsed, 0, sizeof(transfer_elapsed));
-
-			/*
-			 * Calculate the effective start time: the time we actually
-			 * started, plus the total time we spent stopped.
-			 */
-			pv_elapsedtime_add(&init_time, &(info_array[idx].start_time), &(state->signal.toffset));
-
-			/*
-			 * Now get the effective elapsed transfer time - current
-			 * time minus effective start time.
-			 */
-			pv_elapsedtime_subtract(&transfer_elapsed, &cur_time, &init_time);
-
-			if (NULL != info_array[idx].state) {
-				info_array[idx].state->transfer.elapsed_seconds =
-				    pv_elapsedtime_seconds(&transfer_elapsed);
-			}
-
-			if (displayed_lines > 0) {
-				debug("%s", "adding newline");
-				pv_tty_write(state, "\n", 1);
-			}
-
-			if (NULL == info_array[idx].state) {
-				debug("%s %d [%d]: %s / %Ld", "fd", fd, idx, "(null state)", position_now);
 			} else {
-				debug("%s %d [%d]: %Lf / %Ld", "fd", fd, idx,
-				      info_array[idx].state->transfer.elapsed_seconds, position_now);
+				/*
+				 * It this watched item is a single FD, and
+				 * that FD is no longer usable, mark the
+				 * item as finished and move on.
+				 */
+				if ((NULL == watching[watch_idx].info_array) || (0 == watching[watch_idx].array_length)
+				    || (watching[watch_idx].info_array[0].unused)
+				    || (!watching[watch_idx].info_array[0].displayable)) {
+					watching[watch_idx].finished = true;
+					continue;
+				}
 			}
 
-			if (NULL != info_array[idx].state) {
-				info_array[idx].state->transfer.transferred = position_now;
-				info_array[idx].state->transfer.total_written = position_now;
-				pv_display(info_array[idx].state, false);
+			/*
+			 * Run through the info array of this item.
+			 */
+			for (info_idx = 0;
+			     NULL != watching[watch_idx].info_array && info_idx < watching[watch_idx].array_length;
+			     info_idx++) {
+				off_t position_now;
+				pvwatchfd_t info_item = &(watching[watch_idx].info_array[info_idx]);
+
+				/* Skip unused array entries. */
+				if (info_item->unused)
+					continue;
+
+				/* No more lines if we've filled the display. */
+				if (displayed_lines >= (int) (state->control.height))
+					break;
+
+				if (!info_item->displayable) {
+					/*
+					 * Non-displayable fd - just remove if
+					 * changed.
+					 */
+					if (pv_watchfd_changed(info_item)) {
+						debug("%s %d: %s", "fd", info_item->watch_fd,
+						      "non-displayable, and has changed - removing");
+						info_item->unused = true;
+						info_item->displayable = false;
+						pv_freecontents_watchfd(info_item);
+					}
+					continue;
+				}
+
+				if (info_item->watch_fd < 0) {
+					debug("%s %d: %s", "fd", info_item->watch_fd, "negative fd - skipping");
+					continue;
+				}
+
+				/*
+				 * Displayable fd - display, or remove if
+				 * changed.
+				 */
+
+				position_now = -1;
+
+				if (info_item->closed) {
+					/*
+					 * Closed fd - check how long since
+					 * it was closed.
+					 */
+					struct timespec time_since_closed;
+					long double seconds_since_closed;
+
+					memset(&time_since_closed, 0, sizeof(time_since_closed));
+					pv_elapsedtime_subtract(&time_since_closed, &cur_time, &(info_item->end_time));
+					seconds_since_closed = pv_elapsedtime_seconds(&time_since_closed);
+
+					/* Closed for long enough - remove. */
+					if (seconds_since_closed > state->control.interval) {
+						debug("%s %d: %s (%Lf s)", "fd", info_item->watch_fd,
+						      "closed for long enough - removing", seconds_since_closed);
+						info_item->unused = true;
+						info_item->displayable = false;
+						pv_freecontents_watchfd(info_item);
+						continue;
+					}
+
+				} else {
+
+					/*
+					 * Open fd - get its current
+					 * position.
+					 */
+
+					position_now = pv_watchfd_position(info_item);
+
+					if (position_now < 0) {
+						/*
+						 * The fd was closed - mark
+						 * as closed.
+						 */
+						debug("%s %d: %s", "fd", info_item->watch_fd, "marking as closed");
+						pv_elapsedtime_copy(&(info_item->end_time), &cur_time);
+						info_item->closed = true;
+					}
+				}
+
+				if (position_now >= 0) {
+					/*
+					 * If the fd is still open and we
+					 * got its current position, update
+					 * its position and timers.
+					 */
+					struct timespec init_time, transfer_elapsed;
+
+					info_item->position = position_now;
+
+					memset(&init_time, 0, sizeof(init_time));
+					memset(&transfer_elapsed, 0, sizeof(transfer_elapsed));
+
+					/*
+					 * Calculate the effective start
+					 * time: the time we actually
+					 * started, plus the total time we
+					 * spent stopped.
+					 */
+					pv_elapsedtime_add(&init_time, &(info_item->start_time),
+							   &(state->signal.toffset));
+
+					/*
+					 * Now get the effective elapsed
+					 * transfer time - current time
+					 * minus effective start time.
+					 */
+					pv_elapsedtime_subtract(&transfer_elapsed, &cur_time, &init_time);
+
+					info_item->transfer.elapsed_seconds = pv_elapsedtime_seconds(&transfer_elapsed);
+				}
+
+				/*
+				 * Now display the information about the fd.
+				 */
+
+				if (displayed_lines > 0) {
+					debug("%s", "adding newline");
+					pv_tty_write(&(state->flags), "\n", 1);
+				}
+
+				debug("%s %d, %s %d [%d/%d]: %Lf / %Ld", "pid", (int) (info_item->watch_pid), "fd",
+				      info_item->watch_fd, watch_idx, info_idx, info_item->transfer.elapsed_seconds,
+				      info_item->position);
+
+				if (terminal_resized) {
+					pv_watchpid_setname(state, info_item);
+					info_item->flags.reparse_display = 1;
+				}
+
+				info_item->transfer.transferred = info_item->position;
+				info_item->transfer.total_written = info_item->position;
+				state->control.name = info_item->display_name;
+				state->control.size = info_item->size;
+
+				pv_display(&(state->status),
+					   &(state->control), &(info_item->flags),
+					   &(info_item->transfer), &(info_item->calc),
+					   &(state->cursor), &(info_item->display), NULL, false);
+
+				/*@-mustfreeonly@ */
+				state->control.name = NULL;
+				/*
+				 * splint warns of a memory leak, but we'd
+				 * set name to be an alias of display_name,
+				 * so nothing is lost here.
+				 */
+				/*@+mustfreeonly@ */
+
 				displayed_lines++;
 			}
+
 		}
 
 		/*
@@ -1004,10 +1131,10 @@ int pv_watchpid_loop(pvstate_t state)
 		while (blank_lines > 0) {
 			pvdisplay_width_t blank_count;
 			if (displayed_lines > 0)
-				pv_tty_write(state, "\n", 1);
+				pv_tty_write(&(state->flags), "\n", 1);
 			for (blank_count = 0; blank_count < state->control.width; blank_count++)
-				pv_tty_write(state, " ", 1);
-			pv_tty_write(state, "\r", 1);
+				pv_tty_write(&(state->flags), " ", 1);
+			pv_tty_write(&(state->flags), "\r", 1);
 			blank_lines--;
 			displayed_lines++;
 		}
@@ -1015,41 +1142,61 @@ int pv_watchpid_loop(pvstate_t state)
 		debug("%s: %d", "displayed lines", displayed_lines);
 
 		while (displayed_lines > 1) {
-			pv_tty_write(state, "\033[A", 3);
+			pv_tty_write(&(state->flags), "\033[A", 3);
 			displayed_lines--;
 		}
+
+		/* Check whether all watched items have finished. */
+		all_watching_finished = true;
+		for (watch_idx = 0; watch_idx < state->watchfd.count; watch_idx++) {
+			if (watching[watch_idx].finished)
+				continue;
+			all_watching_finished = false;
+			break;
+		}
+	}
+
+	if (!state->control.numeric) {
+		/*
+		 * Move the cursor past the displayed lines, so the last
+		 * output stays on the screen and isn't overwritten by the
+		 * command prompt (#81).
+		 */
+		while (prev_displayed_lines > 0) {
+			pv_tty_write(&(state->flags), "\n", 1);
+			prev_displayed_lines--;
+		}
 	}
 
 	/*
-	 * Clean up our displayed lines on exit.
+	 * If a signal caused the end of the loop, reflect that in the
+	 * return code.
 	 */
-	blank_lines = prev_displayed_lines;
-	while (blank_lines > 0) {
-		pvdisplay_width_t blank_count;
-		for (blank_count = 0; blank_count < state->control.width; blank_count++)
-			pv_tty_write(state, " ", 1);
-		pv_tty_write(state, "\r", 1);
-		blank_lines--;
-		if (blank_lines > 0)
-			pv_tty_write(state, "\n", 1);
-	}
-	while (prev_displayed_lines > 1) {
-		pv_tty_write(state, "\033[A", 3);
-		prev_displayed_lines--;
-	}
+	if (1 == state->flags.trigger_exit)
+		state->status.exit_status |= PV_ERROREXIT_SIGNAL;
 
-	/*
-	 * Free the per-fd state.
-	 */
-	for (idx = 0; NULL != info_array && idx < array_length; idx++) {
-		if (NULL == info_array[idx].state)
+      end_pv_watchfd_loop:
+	/* Free all allocated memory. */
+	/*@-compdestroy@ */
+	for (watch_idx = 0; NULL != watching && watch_idx < state->watchfd.count; watch_idx++) {
+		int info_idx;
+		if (NULL == watching[watch_idx].info_array)
 			continue;
-		pv_state_free(info_array[idx].state);
-		info_array[idx].state = NULL;
+		for (info_idx = 0; info_idx < watching[watch_idx].array_length; info_idx++) {
+			pv_freecontents_watchfd(&(watching[watch_idx].info_array[info_idx]));
+		}
+		free(watching[watch_idx].info_array);
+		watching[watch_idx].info_array = NULL;
+		watching[watch_idx].array_length = 0;
 	}
+	free(watching);
+	watching = NULL;
 
-	if (NULL != info_array)
-		free(info_array);
-
-	return 0;
+	return state->status.exit_status;
+	/*@+compdestroy @ */
+	/*
+	 * splint warns that watching[].info_array->transfer.transfer_buffer
+	 * and other deep structures may not be deallocated and so leak
+	 * memory, but that's what pv_freecontents_watchfd() takes care of.
+	 */
 }
